package checkmk

import (
	"bytes"
	"context"
	"fmt"
	"os"
	"os/exec"
	"path/filepath"
	"strings"
	"time"
)

type LocalCommandBackend struct {
	ctx           context.Context
	timeout       time.Duration
	satelliteName string
}

func NewLocalCommandBackend(ctx context.Context, timeout time.Duration, satelliteName string) *LocalCommandBackend {
	return &LocalCommandBackend{
		ctx:           ctx,
		timeout:       timeout,
		satelliteName: satelliteName,
	}
}

func (l *LocalCommandBackend) RunCommandRaw(host string, directArgs []string, env map[string]string, stdin string) (string, error) {
	ctx, cancel := context.WithTimeout(l.ctx, l.timeout)
	defer cancel()

	c := exec.CommandContext(ctx, directArgs[0], directArgs[1:]...)
	e := make([]string, 0)

	for k, v := range env {
		e = append(e, fmt.Sprintf("%s=%s", k, v))
	}

	c.Env = e

	finished := make(chan error)
	output := &bytes.Buffer{}
	c.Stdout = output
	c.Stderr = output

	cmdArgs := strings.Join(directArgs, " ")

	err := c.Start()
	if err != nil {
		return "", &Error{HostUUID: host, Message: fmt.Sprint(cmdArgs, " => could not start process"), Err: err}
	}

	go func() {
		err := c.Wait()
		select {
		case <-ctx.Done():
			return
		default:
			finished <- err
		}
	}()

	select {
	case err := <-finished:
		out := output.String()
		if err != nil {
			return out, &Error{Name: l.satelliteName, HostUUID: host, Message: fmt.Sprint(cmdArgs, " => \"", out, "\""), Err: err}
		}
		return out, nil
	case <-ctx.Done():
		return "", &Error{Name: l.satelliteName, HostUUID: host, Message: fmt.Sprint(cmdArgs, " => process timed out")}
	}
}

func (l *LocalCommandBackend) RunCommand(hostuuid string, args []string, env map[string]string) (string, error) {
	return l.RunCommandRaw(hostuuid, args, env, "")
}

func (l *LocalCommandBackend) WriteFile(hostuuid, path string, content []byte) error {
	d := filepath.Dir(path)
	err := os.MkdirAll(d, 0O0755)
	if err != nil {
		return &Error{Name: l.satelliteName, HostUUID: hostuuid, Message: fmt.Sprint("could not create directory ", d), Err: err}
	}
	fl, err := os.Create(path)
	if err != nil {
		return &Error{Name: l.satelliteName, HostUUID: hostuuid, Message: fmt.Sprint("could not open file ", path, " for writing"), Err: err}
	}
	defer fl.Close()
	_, err = fl.Write(content)
	if err != nil {
		return &Error{Name: l.satelliteName, HostUUID: hostuuid, Message: fmt.Sprint("could not write content of file ", path), Err: err}
	}
	return nil
}
