package cli

import (
	"context"
	"flag"
	"io/ioutil"
	"log"
	"os"
	"time"

	"github.com/it-novum/nsta/nstaclient"
	"github.com/it-novum/nsta/transfer"
	"gopkg.in/ini.v1"
)

// Cli holds the information about command line parameters
type Cli struct {
	configPath         string
	logFile            string
	accessLogFile      string
	gearmanAddress     string
	listenAddressHTTPS string
	listenAddressHTTP  string
	mode               string
	apiKey             string
	endpoint           string
	satelliteID        int
	tlsKey             string
	tlsCert            string
	mycnf              string
	nagiosPath         string
	insecureHTTPS      bool
	proxy              string
	interval           int
	disableInstantPush bool
}

func (c *Cli) readConfig() error {
	cfg, err := ini.Load(c.configPath)
	if err != nil {
		return err
	}
	nstaSec, err := cfg.GetSection("nsta")
	if err != nil {
		return err
	}
	if key, err := nstaSec.GetKey("gearman"); err == nil {
		c.gearmanAddress = key.String()
	} else {
		c.gearmanAddress = "127.0.0.1:4730"
	}
	if key, err := nstaSec.GetKey("listen-http"); err == nil {
		c.listenAddressHTTP = key.String()
	} else {
		// we use ipv4 localhost because some systems have ivp6 disabled ...
		c.listenAddressHTTP = "127.0.0.1:7473"
	}
	if key, err := nstaSec.GetKey("listen-https"); err == nil {
		c.listenAddressHTTPS = key.String()
	} else {
		c.listenAddressHTTPS = ":7474"
	}
	if key, err := nstaSec.GetKey("tls-key"); err == nil {
		c.tlsKey = key.String()
	} else {
		c.tlsKey = "/etc/ssl/private/ssl-cert-snakeoil.key"
	}
	if key, err := nstaSec.GetKey("tls-cert"); err == nil {
		c.tlsCert = key.String()
	} else {
		c.tlsCert = "/etc/ssl/certs/ssl-cert-snakeoil.pem"
	}
	if key, err := nstaSec.GetKey("mode"); err == nil {
		c.mode = key.String()
	} else {
		c.mode = "server"
	}
	if key, err := nstaSec.GetKey("api-key"); err == nil {
		c.apiKey = key.String()
	} else {
		c.apiKey = ""
	}
	if key, err := nstaSec.GetKey("endpoint"); err == nil {
		c.endpoint = key.String()
	} else {
		c.endpoint = ""
	}
	if key, err := nstaSec.GetKey("satellite-id"); err == nil {
		c.satelliteID = key.MustInt(-1)
	} else {
		c.satelliteID = -1
	}
	if key, err := nstaSec.GetKey("mycnf"); err == nil {
		c.mycnf = key.String()
	} else {
		c.mycnf = "/opt/openitc/etc/mysql/mysql.cnf"
	}
	if key, err := nstaSec.GetKey("nagios-path"); err == nil {
		c.nagiosPath = key.String()
	} else {
		c.nagiosPath = "/opt/openitc/nagios"
	}
	if key, err := nstaSec.GetKey("insecure-https"); err == nil {
		c.insecureHTTPS = key.MustBool(false)
	} else {
		c.insecureHTTPS = false
	}
	if key, err := nstaSec.GetKey("proxy"); err == nil {
		c.proxy = key.String()
	} else {
		c.proxy = ""
	}
	if key, err := nstaSec.GetKey("log-file"); err == nil {
		c.logFile = key.String()
	} else {
		c.logFile = "-"
	}
	if key, err := nstaSec.GetKey("access-log-file"); err == nil {
		c.accessLogFile = key.String()
	} else {
		c.accessLogFile = "-"
	}
	if key, err := nstaSec.GetKey("interval"); err == nil {
		c.interval = key.MustInt(5)
	} else {
		c.interval = 5
	}
	if key, err := nstaSec.GetKey("disable-instant-push"); err == nil {
		c.disableInstantPush = key.MustBool(false)
	} else {
		c.disableInstantPush = false
	}
	return nil
}

// ParseCLI uses flag to parse the command line parameters
func ParseCLI() *Cli {
	c := &Cli{}
	flag.StringVar(&c.configPath, "config", "/opt/openitc/etc/nsta/nsta.ini", "Path to nsta.ini configuration")
	flag.Parse()

	if err := c.readConfig(); err != nil {
		log.Fatalln(err)
	}

	if c.mode == "https_pull" {
		if c.apiKey == "" {
			log.Fatalln("missing -api-key")
		}
		if c.satelliteID == -1 {
			log.Fatalln("missing -satellite-id")
		}
	}
	if c.mode == "https_push" {
		if c.apiKey == "" {
			log.Fatalln("missing -api-key")
		}
		if c.satelliteID == -1 {
			log.Fatalln("missing -satellite-id")
		}
		if c.endpoint == "" {
			log.Fatalln("missing -endpoint url")
		}
	}

	return c
}

func Run() {
	var err error
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	cli := ParseCLI()

	accessLogWriter := ioutil.Discard
	if cli.accessLogFile != "" {
		if cli.accessLogFile == "-" {
			accessLogWriter = os.Stdout
		} else {
			accessLogWriter, err = os.OpenFile(cli.accessLogFile, os.O_APPEND|os.O_CREATE, os.FileMode(0o644))
			if err != nil {
				log.Println(err)
				os.Exit(1)
			}
		}
	}

	logWriter := ioutil.Discard
	if cli.logFile != "" {
		if cli.logFile == "-" {
			logWriter = os.Stdout
		} else {
			logWriter, err = os.OpenFile(cli.logFile, os.O_APPEND|os.O_CREATE, os.FileMode(0o644))
			if err != nil {
				log.Println(err)
				os.Exit(1)
			}
		}
	}

	log.SetOutput(logWriter)

	if cli.mode == "server" {
		m, err := transfer.New(ctx, accessLogWriter, cli.mycnf, cli.gearmanAddress, cli.listenAddressHTTP, cli.listenAddressHTTPS, cli.tlsCert, cli.tlsKey, cli.nagiosPath)
		if err != nil {
			log.Fatalln(err)
		}
		m.Run()
	} else if cli.mode == "https_push" {
		client, err := nstaclient.NewPushClient(ctx, cli.gearmanAddress, cli.endpoint, cli.satelliteID, cli.apiKey, cli.nagiosPath, cli.proxy, !cli.insecureHTTPS, time.Second*time.Duration(cli.interval), !cli.disableInstantPush)
		if err != nil {
			log.Fatalln(err)
		}
		client.Run()
	} else if cli.mode == "https_pull" {
		client, err := nstaclient.NewPullClient(ctx, accessLogWriter, cli.gearmanAddress, cli.listenAddressHTTP, cli.listenAddressHTTPS, cli.tlsCert, cli.tlsKey, cli.satelliteID, cli.apiKey, cli.nagiosPath)
		if err != nil {
			log.Fatalln(err)
		}
		client.Run()
	}
}
