package gearman

import (
	"context"
	"fmt"
	"log"
	"sync"
	"time"

	"github.com/it-novum/gearman-go/client"
	"github.com/it-novum/gearman-go/worker"
)

// Error implements a gearman client specific error
type Error struct {
	clientName string
	msg        string
	err        error
}

// Unwrap implements errors.Unwrap
func (e *Error) Unwrap() error {
	return e.err
}

// Unwrap implements errors.Error
func (e *Error) Error() string {
	if e.err != nil {
		if e.msg != "" {
			return fmt.Sprintf("[%s] gearman %s: %s", e.clientName, e.msg, e.err.Error())
		}
		return fmt.Sprintf("[%s] gearman client error: %s", e.clientName, e.err.Error())
	}
	return fmt.Sprintf("[%s] %s", e.clientName, e.msg)
}

// ConnectionStatus is an enum holding the current status of the connection
type ConnectionStatus int

const (
	// Connected is returend if the client has an active connection to the server
	Connected ConnectionStatus = 0
	// Disconnected is returend if the client can't connect to the server
	Disconnected ConnectionStatus = 1
	// Stopped is returend if the client has been closed
	Stopped ConnectionStatus = 2
)

type clientType int

const (
	master    clientType = 0
	satellite clientType = 1
)

// Client represents the connection to gearman server
type Client struct {
	ctx                   context.Context
	ctxcancel             context.CancelFunc
	mtx                   sync.Mutex
	wg                    sync.WaitGroup
	stopped               bool
	workerDisconnectError *worker.WorkerDisconnectError

	clientType clientType
	clientName string

	worker  *worker.Worker
	client  *client.Client
	address string

	messageSend    SyncMessageFIFO
	messageReceive chan<- *Message

	statusClient ConnectionStatus
	statusWorker ConnectionStatus
}

// IsRunning returns true if gearman worker and client is running
func (c *Client) IsRunning() bool {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.worker != nil && c.client != nil && c.statusClient == Connected && c.statusWorker == Connected {
		return true
	}
	return false
}

func newClient(ctx context.Context, clientType clientType, clientName, address string) *Client {
	ctx, cancel := context.WithCancel(ctx)
	c := &Client{
		ctx:        ctx,
		ctxcancel:  cancel,
		clientType: clientType,
		clientName: clientName,

		worker:  nil,
		client:  nil,
		address: address,

		stopped:               false,
		workerDisconnectError: nil,

		statusClient: Disconnected,
		statusWorker: Disconnected,
	}
	c.messageSend.Start()
	c.wg.Add(1)

	return c
}

// NewMaster creates a gearman client connection at the oitc master
func NewMaster(ctx context.Context, clientName, address string, messageReceive chan<- *Message) *Client {
	c := newClient(ctx, master, clientName, address)
	c.messageReceive = messageReceive

	return c
}

// NewSatellite creates a gearman client connection to a oitc satellite server
func NewSatellite(ctx context.Context, clientName, address string, messageReceive chan<- *Message) *Client {
	c := newClient(ctx, satellite, clientName, address)
	c.messageReceive = messageReceive

	return c
}

func (c *Client) connectWorker() {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if !c.stopped {
		if c.worker == nil {
			log.Println(&Error{clientName: c.clientName, msg: "connect to gearman server as worker"})
			w := worker.New(worker.OneByOne)

			if err := w.AddServer("tcp", c.address); err != nil {
				log.Println(&Error{clientName: c.clientName, err: err})
				return
			}

			if c.clientType == master {
				if err := w.AddFunc("statusngin_cmd_sattx", func(job worker.Job) ([]byte, error) {
					c.messageReceive <- &Message{Type: MessageCommandSatTX, String: string(job.Data())}
					return nil, nil
				}, worker.Unlimited); err != nil {
					log.Println(&Error{clientName: c.clientName, err: err})
					return
				}
				if err := w.AddFunc("oitc_checkmk_sattx", func(job worker.Job) ([]byte, error) {
					c.messageReceive <- &Message{Type: MessageCheckmk, String: string(job.Data())}
					return nil, nil
				}, worker.Unlimited); err != nil {
					log.Println(&Error{clientName: c.clientName, err: err})
					return
				}
			} else {
				if err := w.AddFunc("statusngin_ochp", func(job worker.Job) ([]byte, error) {
					c.messageReceive <- &Message{Type: MessageOCHP, String: string(job.Data())}
					return nil, nil
				}, worker.Unlimited); err != nil {
					log.Println(&Error{clientName: c.clientName, err: err})
					return
				}
				if err := w.AddFunc("statusngin_ocsp", func(job worker.Job) ([]byte, error) {
					c.messageReceive <- &Message{Type: MessageOCSP, String: string(job.Data())}
					return nil, nil
				}, worker.Unlimited); err != nil {
					log.Println(&Error{clientName: c.clientName, err: err})
					return
				}
				if err := w.AddFunc("phpnsta_custom_input", func(job worker.Job) ([]byte, error) {
					c.messageReceive <- &Message{Type: MessageCustomInput, String: string(job.Data())}
					return nil, nil
				}, worker.Unlimited); err != nil {
					log.Println(&Error{clientName: c.clientName, err: err})
					return
				}
				if err := w.AddFunc("statusngin_cmd_mastertx", func(job worker.Job) ([]byte, error) {
					c.messageReceive <- &Message{Type: MessageCommandMasterTX, String: string(job.Data())}
					return nil, nil
				}, worker.Unlimited); err != nil {
					log.Println(&Error{clientName: c.clientName, err: err})
					return
				}
			}

			w.ErrorHandler = func(err error) {
				c.mtx.Lock()
				defer c.mtx.Unlock()

				// if it is already disconnected we get into a reconnect loop
				if c.statusWorker == Connected {
					disconnect, ok := err.(*worker.WorkerDisconnectError)
					if ok {
						c.statusWorker = Disconnected
						c.workerDisconnectError = disconnect
					}
				}
			}

			if err := w.Ready(); err != nil {
				log.Println(&Error{clientName: c.clientName, err: err, msg: "connection failed"})
				return
			}
			log.Println(&Error{clientName: c.clientName, msg: "connection to gearman successful"})

			go w.Work()
			c.worker = w
			c.statusWorker = Connected
		} else if c.statusWorker == Disconnected {
			log.Println(&Error{clientName: c.clientName, msg: "reconnect required"})
			c.workerDisconnectError.Reconnect()
			c.statusWorker = Connected
			c.workerDisconnectError = nil
		}
	}
}

func (c *Client) handleClientError(err error) {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.statusClient == Connected {
		log.Println(&Error{clientName: c.clientName, err: err, msg: "connection failed"})
		c.statusClient = Disconnected
		c.client.Close()
		c.client = nil
	}
}

func (c *Client) connectClient() {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.client == nil {
		log.Println(&Error{clientName: c.clientName, msg: "connect to gearman server as client"})
		gc, err := client.New("tcp", c.address)
		if err != nil {
			log.Println(&Error{clientName: c.clientName, err: err, msg: "connection failed"})
			return
		}

		gc.ErrorHandler = c.handleClientError

		c.client = gc
		c.statusClient = Connected
	} else if c.statusClient == Disconnected {
		c.statusClient = Connected
	}
}

// Run starts the event loop
func (c *Client) Run() {
	ticker := time.NewTicker(time.Second * 5)
	defer func() {
		ticker.Stop()
		if c.worker != nil {
			log.Println(&Error{clientName: c.clientName, msg: "close gearman worker connection"})
			c.worker.Close()
		}
		if c.client != nil {
			log.Println(&Error{clientName: c.clientName, msg: "close gearman client connection"})
			c.client.Close()
		}
		log.Println(&Error{clientName: c.clientName, msg: "close gearman fifo"})
		c.messageSend.Close()
		c.statusWorker = Stopped
		c.statusClient = Stopped
		c.wg.Done()
	}()

	c.connectWorker()
	c.connectClient()

	for {
		if c.IsRunning() {
			select {
			case <-c.ctx.Done():
				return
			case <-ticker.C:
				c.connectWorker()
				c.connectClient()
			case <-c.messageSend.Notify():
				// retest again if we are still running
				if c.IsRunning() {
					msgs := c.messageSend.GetAll(16 * 1024)
					for i, msg := range msgs {
						if err := c.sendMessage(msg); err != nil {
							c.handleClientError(err)
							c.messageSend.PushAll(msgs[i:], true)
							break
						}
					}
				}
			}
		} else {
			log.Println(&Error{clientName: c.clientName, msg: "Waiting for gearman reconnect"})
			select {
			case <-c.ctx.Done():
				return
			case <-ticker.C:
				c.connectWorker()
				c.connectClient()
			}
		}
	}
}

// Close should be called before programms exit
func (c *Client) Close() {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if !c.stopped {
		c.stopped = true
		c.ctxcancel()
	}
	c.wg.Wait()
}

func (c *Client) sendOCHP(msg string) error {
	_, err := c.client.DoBg("statusngin_worker_ochp", []byte(msg), client.JobNormal)
	if err != nil {
		return &Error{clientName: c.clientName, err: err, msg: "could not send ochp to server"}
	}
	return nil
}

func (c *Client) sendOCSP(msg string) error {
	_, err := c.client.DoBg("statusngin_worker_ocsp", []byte(msg), client.JobNormal)
	if err != nil {
		return &Error{clientName: c.clientName, err: err, msg: "could not send ocsp to server"}
	}
	return nil
}

func (c *Client) sendCommand(msg string) error {
	_, err := c.client.DoBg("statusngin_cmd", []byte(msg), client.JobNormal)
	if err != nil {
		return &Error{clientName: c.clientName, err: err, msg: "could not send command to server"}
	}
	return nil
}

func (c *Client) sendCustomInput(msg string) error {
	_, err := c.client.DoBg("phpnsta_custom_data", []byte(msg), client.JobNormal)
	if err != nil {
		return &Error{clientName: c.clientName, err: err, msg: "could not send custom input to server"}
	}
	return nil
}

func (c *Client) sendOITCGearman(msg string) error {
	_, err := c.client.DoBg("oitc_gearman", []byte(msg), client.JobNormal)
	if err != nil {
		return &Error{clientName: c.clientName, err: err, msg: "could not send gearman worker job to server"}
	}
	return nil
}

func (c *Client) sendMessage(msg *Message) error {
	switch msg.Type {
	case MessageOCHP:
		return c.sendOCHP(msg.String)
	case MessageOCSP:
		return c.sendOCSP(msg.String)
	case MessageCommandMasterTX:
		return c.sendCommand(msg.String)
	case MessageCommandSatTX:
		return c.sendCommand(msg.String)
	case MessageCustomInput:
		return c.sendCustomInput(msg.String)
	case MessageCheckmk:
		return c.sendOITCGearman(msg.String)
	}
	return &Error{clientName: c.clientName, msg: "unknown message type"}
}

// SendMessage queues a message for sending to gearman
func (c *Client) SendMessage(msg *Message) {
	c.messageSend.Push(msg)
}
