package gearman

import (
	"context"
	"fmt"
	"testing"
	"time"
)

func TestNewMaster(t *testing.T) {
	msgReceive := make(chan *Message, 200)
	defer close(msgReceive)
	master := NewMaster(context.Background(), "test-master", "127.0.0.1:4730", msgReceive)
	go master.Run()
	master.Close()
}

func TestNewSatellite(t *testing.T) {
	msgReceive := make(chan *Message, 200)
	defer close(msgReceive)

	sat := NewSatellite(context.Background(), "test-sat", "127.0.0.1:4730", msgReceive)
	go sat.Run()
	defer sat.Close()
	for {
		timeout := time.NewTimer(time.Second * 10)
		defer timeout.Stop()
		select {
		case <-timeout.C:
			t.Fatal("Timeout during startup of satellite")
		default:
			if sat.IsRunning() {
				return
			}
			time.Sleep(time.Millisecond * 50)
		}
	}
}

func TestCancel(t *testing.T) {
	msgReceive := make(chan *Message, 200)
	defer close(msgReceive)
	ctx, cancel := context.WithCancel(context.Background())
	master := NewMaster(ctx, "test-master", "127.0.0.1:4730", msgReceive)
	go master.Run()

	cancel()
	for {
		timeout := time.NewTimer(time.Second * 10)
		select {
		case <-timeout.C:
			t.Fatal("Timeout during shutdown of master")
		default:
			if master.stopped {
				return
			}
			time.Sleep(time.Millisecond * 50)
		}
	}
}

func TestError(t *testing.T) {
	errMap := map[string]error{
		"[test] test":                       &Error{clientName: "test", msg: "test"},
		"[test] gearman test: test":         &Error{clientName: "test", err: fmt.Errorf("test"), msg: "test"},
		"[test] gearman client error: test": &Error{clientName: "test", err: fmt.Errorf("test")},
	}

	for expected, err := range errMap {
		if fmt.Sprint(err) != expected {
			t.Error("expected", expected, "but got", fmt.Sprint(err))
		}
	}
}
