package gearman

import (
	"log"
	"time"

	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

// MessageType identifes the type of a statusengine message
type MessageType int

const (
	// MessageOCHP is an OCHP statusengine message
	MessageOCHP MessageType = 0
	// MessageOCSP is an OCSP statusengine message
	MessageOCSP MessageType = 1
	// MessageCommandSatTX is a command statusengine message
	MessageCommandSatTX MessageType = 2
	// MessageCustomInput is a phpnsta custom message transmission
	MessageCustomInput MessageType = 3
	// MessageCommandMasterTX transfers commands from sat to master
	MessageCommandMasterTX MessageType = 4
	// MessageCheckmk transfers Checkmk commands from master to sat
	MessageCheckmk MessageType = 5
	// MessageTimesync signals a timesync to satellite
	MessageTimesync MessageType = 6
)

// MessageContentTX represents a typical otic gearman message
type MessageContentTX struct {
	SatelliteID int
	Command     string
	Messages    []*MessageContentTX
}

// Message holds the information of a message for queuing
type Message struct {
	Type   MessageType
	String string
}

// Content parses the content of String
func (m *Message) Content() *MessageContentTX {
	c := &MessageContentTX{}
	err := json.Unmarshal([]byte(m.String), c)
	if err != nil {
		log.Println("could not parse gearman content: ", err)
		return nil
	}
	return c
}

// CheckmkDataMessage represents the information for checkmk scan
type CheckmkDataMessage struct {
	ScanID   int
	Hostuuid string
	Task     string
	File     string
	FilePath string
}

// CheckmkMessage is a special message for checkmk related commands
type CheckmkMessage struct {
	SatelliteID int
	Command     string
	Data        CheckmkDataMessage
}

// CheckmkResponse holds the information for the response to the gearman worker
type CheckmkResponse struct {
	Task             string `json:"task"`
	SatelliteID      int
	Command          string
	ScanID           int
	CheckTypesResult string
	DumpHostResult   string
	RawInfoResult    string
	Error            string
}

// TimesyncMessage represents a timesync message
type TimesyncMessage struct {
	SatelliteID int
	SetTime     time.Time
}

// Serialize returns the php representation of the response message, required by gearman_worker
func (c *CheckmkResponse) Serialize() (string, error) {
	c.Task = "CheckmkSatResult"
	r, err := json.Marshal(c)
	if err != nil {
		return "", err
	}
	return string(r), nil
}
