package nstaclient

import (
	"context"
	"fmt"
	"io"
	"log"
	"os"
	"os/signal"
	"path/filepath"
	"sync"
	"syscall"
	"time"

	"github.com/it-novum/nsta/checkmk"
	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpproto"
	"github.com/it-novum/nsta/httpserver"
	"github.com/it-novum/nsta/nstautil"
	"github.com/it-novum/nsta/satstatus"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

type ClientPull struct {
	ctx       context.Context
	ctxcancel context.CancelFunc
	wg        sync.WaitGroup

	satelliteID     int
	satelliteAPIKey string

	gearman          *gearman.Client
	webserver        *httpserver.NSTAHTTPServer
	nagiosPath       string
	nagiosConfigPath string
	toLocalSat       gearman.SyncMessageFIFO
	toMaster         gearman.SyncMessageFIFO
	cmk              checkmk.Checkmk
	cmdBackend       checkmk.CommandBackend

	lastExport time.Time
}

func (c *ClientPull) SatelliteID() int {
	return c.satelliteID
}

func (c *ClientPull) SatelliteName() string {
	return "Local"
}

func (c *ClientPull) APIKey() string {
	return c.satelliteAPIKey
}

func (c *ClientPull) Incoming(msg *gearman.Message) {
	c.toLocalSat.Push(msg)
}

func (c *ClientPull) Outgoing() []*gearman.Message {
	return c.toMaster.GetAll(8192)
}

func (c *ClientPull) LastExport() time.Time {
	return c.lastExport
}

func (c *ClientPull) Export(msg *httpproto.ExportData) error {
	err := nstautil.InstallExport(c.nagiosConfigPath, msg.TARBase64)
	if err == nil {
		c.lastExport = msg.LastExport
	}
	return err
}

// WebStatus is only required on the master, but implements SatelliteIO
func (c *ClientPull) WebStatus(webStatus *satstatus.Status) {
}

func NewPullClient(ctx context.Context, accessLogWriter io.Writer, addr, webserverAddressHTTP, webserverAddressHTTPS, certFile, keyFile string, satelliteID int, apiKey string, nagiosPath string) (*ClientPull, error) {
	ctx, ctxcancel := context.WithCancel(ctx)
	c := &ClientPull{
		ctx:              ctx,
		ctxcancel:        ctxcancel,
		satelliteID:      satelliteID,
		satelliteAPIKey:  apiKey,
		webserver:        httpserver.New(accessLogWriter, nagiosPath, webserverAddressHTTP, webserverAddressHTTPS, certFile, keyFile, httpserver.WebserverTypeSatellite),
		lastExport:       time.Time{},
		nagiosPath:       nagiosPath,
		nagiosConfigPath: filepath.Join(nagiosPath, "etc"),
		cmdBackend:       checkmk.NewLocalCommandBackend(ctx, time.Second*120, "Local"),
	}
	c.cmk.SatelliteName = "Local"
	c.toLocalSat.Start()
	c.toMaster.Start()
	c.gearman = gearman.NewSatellite(ctx, "Local", addr, c.toMaster.Direct())
	c.webserver.RegisterSatelliteIO(c)
	c.wg.Add(1)

	go c.gearman.Run()
	c.webserver.Run()

	return c, nil
}

func (c *ClientPull) Stop() {
	c.ctxcancel()
	c.wg.Wait()
}

func (c *ClientPull) Run() {
	defer func() {
		c.webserver.Stop()
		c.webserver.UnregisterSatelliteIO(c.SatelliteID())
		c.toLocalSat.Close()
		c.toMaster.Close()
		c.wg.Done()
	}()

	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM)

	for {
		select {
		case <-c.ctx.Done():
			return
		case <-sigs:
			c.ctxcancel()
		case <-c.toLocalSat.Notify():
			msg := c.toLocalSat.Get()
			if msg != nil {
				switch msg.Type {
				case gearman.MessageCheckmk:
					go func() {
						response, err := c.cmk.ProcessMessage(c.cmdBackend, msg).Serialize()
						if err != nil {
							log.Println("could no serialize checmk answer:", err)
						}
						c.toMaster.Push(&gearman.Message{
							Type:   gearman.MessageCheckmk,
							String: response,
						})
					}()
				case gearman.MessageTimesync:
					go func() {
						ts := gearman.TimesyncMessage{}
						if err := json.Unmarshal([]byte(msg.String), &ts); err != nil {
							log.Println(&Error{Name: "Local", Message: "received invalid timesync message", Err: err})
						} else {
							if _, err := c.cmdBackend.RunCommandRaw("Local", []string{"sudo", "date", "-s", fmt.Sprint("@", ts.SetTime.Unix())}, map[string]string{}, ""); err != nil {
								log.Println(err)
							} else {
								log.Println(&Error{Name: "Local", Message: "successful updated time"})
							}
						}
					}()
				default:
					c.gearman.SendMessage(msg)
				}
			}
		}
	}
}
