package nstaclient

import (
	"context"
	"fmt"
	"log"
	"os"
	"os/signal"
	"path/filepath"
	"sync"
	"syscall"
	"time"

	"github.com/it-novum/nsta/checkmk"
	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpclient"
	"github.com/it-novum/nsta/nstautil"
)

type ClientPush struct {
	ctx       context.Context
	ctxcancel context.CancelFunc
	wg        sync.WaitGroup

	gearman          *gearman.Client
	client           *httpclient.NSTAHTTPClient
	nagiosPath       string
	nagiosConfigPath string
	cmk              checkmk.Checkmk
	cmdBackend       checkmk.CommandBackend
	interval         time.Duration
	instantPush      bool

	lastExport time.Time

	toMaster gearman.SyncMessageFIFO
}

func NewPushClient(ctx context.Context, addr string, endpoint string, satelliteID int, apiKey string, nagiosPath string, proxy string, verifyCertificate bool, interval time.Duration, instantPush bool) (*ClientPush, error) {
	ctx, ctxcancel := context.WithCancel(ctx)
	_ = ctxcancel
	httpClient, err := httpclient.New(ctx, endpoint, satelliteID, apiKey, proxy, verifyCertificate)
	if err != nil {
		return nil, err
	}
	c := &ClientPush{
		ctx:              ctx,
		ctxcancel:        ctxcancel,
		client:           httpClient,
		lastExport:       time.Time{},
		nagiosPath:       nagiosPath,
		nagiosConfigPath: filepath.Join(nagiosPath, "etc"),
		cmdBackend:       checkmk.NewLocalCommandBackend(ctx, time.Second*120, "Local"),
		interval:         interval,
		instantPush:      instantPush,
	}
	c.cmk.SatelliteName = "Local"
	c.toMaster.Start()
	c.gearman = gearman.NewSatellite(ctx, "Local", addr, c.toMaster.Direct())
	c.wg.Add(1)

	go c.gearman.Run()

	return c, nil
}

func (c *ClientPush) Stop() {
	c.ctxcancel()
	c.wg.Wait()
}

func (c *ClientPush) processMessage() error {
	msgs := c.toMaster.GetAll(8192)
	response, err := c.client.SendMessages(msgs, c.lastExport)
	if err != nil {
		c.toMaster.PushAll(msgs, true)
		log.Println(&Error{Name: "Local", Err: err})
		return err
	}
	if response != nil {
		log.Println(&Error{Name: "Local", Message: fmt.Sprint("sent ", len(msgs), " and received ", len(response.Messages), " messages")})

		for _, r := range response.Messages {
			switch r.Type {
			case gearman.MessageCheckmk:
				go func() {
					response, err := c.cmk.ProcessMessage(c.cmdBackend, r).Serialize()
					if err != nil {
						log.Println(&Error{Name: "Local", Err: err})
						return
					}
					c.toMaster.Push(&gearman.Message{
						Type:   gearman.MessageCheckmk,
						String: response,
					})
				}()
			case gearman.MessageTimesync:
				go func() {
					ts := gearman.TimesyncMessage{}
					if err := json.Unmarshal([]byte(r.String), &ts); err != nil {
						log.Println(&Error{Name: "Local", Message: "received invalid timesync message", Err: err})
					} else {
						if _, err := c.cmdBackend.RunCommandRaw("Local", []string{"sudo", "date", "-s", fmt.Sprint("@", ts.SetTime.Unix())}, map[string]string{}, ""); err != nil {
							log.Println(err)
						} else {
							log.Println(&Error{Name: "Local", Message: "successful updated time"})
						}
					}
				}()
			default:
				c.gearman.SendMessage(r)
			}
		}
		if !response.LastExport.IsZero() {
			if response.LastExport.After(c.lastExport) {
				export, err := c.client.ExportExchange(c.lastExport, "")
				if err == nil && export != nil {
					if export.TARBase64 == "" {
						log.Println(&Error{Name: "Local", Message: "requst new export"})
					} else {
						if err = nstautil.InstallExport(c.nagiosConfigPath, export.TARBase64); err != nil {
							log.Println(&Error{Name: "Local", Err: err})
						} else {
							c.lastExport = export.LastExport
						}

					}
				}
			}
		}
	} else if err != nil {
		log.Println(&Error{Name: "Local", Message: fmt.Sprint("sent ", len(msgs), " messages")})
	}

	return nil
}

func (c *ClientPush) Run() {
	var notify chan struct{}
	emptyChan := make(chan struct{})

	defer func() {
		c.toMaster.Close()
		close(emptyChan)
		c.wg.Done()
	}()

	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM)

	ticker := time.NewTicker(c.interval)
	lastError := false

	for {
		if !c.instantPush || lastError {
			notify = emptyChan
		} else {
			notify = c.toMaster.Notify()
		}
		select {
		case <-c.ctx.Done():
			return
		case <-sigs:
			c.ctxcancel()
			return
		case <-notify:
			err := c.processMessage()
			if err != nil {
				lastError = true
			} else {
				lastError = false
			}
		case <-ticker.C:
			err := c.processMessage()
			if err != nil {
				lastError = true
			} else {
				lastError = false
			}
		}
	}
}
