package nstautil

import (
	"archive/tar"
	"bytes"
	"compress/gzip"
	"encoding/base64"
	"fmt"
	"io"
	"log"
	"os"
	"os/exec"
	"path/filepath"
	"reflect"
	"strings"
)

// TarPackEncode packs a directory to a tar archive, gzip compressed and base64 encoded
func TarPackEncode(directory string, use64 bool) (string, error) {
	var b bytes.Buffer
	var encoder io.WriteCloser
	var compressor *gzip.Writer
	if use64 {
		encoder = base64.NewEncoder(base64.StdEncoding, &b)
		compressor = gzip.NewWriter(encoder)
	} else {
		compressor = gzip.NewWriter(&b)
	}

	tarfile := tar.NewWriter(compressor)
	err := filepath.Walk(directory, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			return err
		}

		if !info.Mode().IsRegular() && !info.Mode().IsDir() {
			return nil
		}

		header, err := tar.FileInfoHeader(info, info.Name())
		if err != nil {
			return err
		}

		header.Name = strings.TrimPrefix(strings.Replace(path, directory, "", -1), string(filepath.Separator))

		if err := tarfile.WriteHeader(header); err != nil {
			return err
		}

		if !info.Mode().IsDir() {
			f, err := os.Open(path)
			if err != nil {
				return err
			}

			if _, err := io.Copy(tarfile, f); err != nil {
				return err
			}
			f.Close()
		}
		return nil
	})
	if err != nil {
		return "", fmt.Errorf("could not create tar archive: %s", err)
	}
	tarfile.Close()
	compressor.Close()
	if use64 {
		encoder.Close()
	}
	return b.String(), nil
}

// TarExtractDecode unpacks a tar archive, gzip compressed and base64 encoded to a destination directory
func TarExtractDecode(directory, content string) error {
	decoder := base64.NewDecoder(base64.StdEncoding, strings.NewReader(content))
	gzip, err := gzip.NewReader(decoder)
	if err != nil {
		return fmt.Errorf("could not decode export file %s", err)
	}
	tarfile := tar.NewReader(gzip)
	for {
		hdr, err := tarfile.Next()
		if err == io.EOF {
			break
		} else if err != nil {
			return fmt.Errorf("could not extract tar file: %s", err)
		}

		switch hdr.Typeflag {
		case tar.TypeDir:
			absPath := filepath.Join(directory, strings.TrimLeft(hdr.Name, "/"))
			os.MkdirAll(absPath, hdr.FileInfo().Mode())
		case tar.TypeReg:
			absPath := filepath.Join(directory, strings.TrimLeft(hdr.Name, "/"))
			fl, err := os.Create(absPath)
			if err != nil {
				return fmt.Errorf("could not extract file %s from tar: %s", absPath, err)
			}
			_, err = io.Copy(fl, tarfile)
			if err != nil {
				return fmt.Errorf("could not extract file %s from tar: %s", absPath, err)
			}
			// we ignore errors here on purpose
			fl.Chmod(hdr.FileInfo().Mode())
			err = fl.Close()
			if err != nil {
				return fmt.Errorf("could not extract file %s from tar: %s", absPath, err)
			}
		default:
			log.Println("unknown file type in tar")
		}
	}
	return nil
}

// InstallExport extracts an export archive and restarts the monitoring system
func InstallExport(basedir, tardata string) error {
	log.Println(&Error{Name: "Local", Message: "remove old configuration"})
	if _, err := os.Stat(filepath.Join(basedir, "config")); !os.IsNotExist(err) {
		if err := os.RemoveAll(filepath.Join(basedir, "config")); err != nil {
			log.Println(&Error{Name: "Local", Err: err})
		}
	}
	log.Println(&Error{Name: "Local", Message: "unpacking new config"})
	err := TarExtractDecode(basedir, tardata)
	if err != nil {
		return err
	}

	log.Println(&Error{Name: "Local", Message: "restart monitoring system"})
	err = exec.Command("/usr/bin/sudo", "/opt/openitc/nagios/bin/restart-monitoring.sh").Run()
	if err != nil {
		log.Println(&Error{Name: "Local", Message: "could not restart monitoring", Err: err})
	}
	return nil
}

// CreateExportTar creates an tar export archive for given satelliteID
func CreateExportTar(nagiosPath string, satelliteID int, use64 bool) string {
	satPath := fmt.Sprintf("%s/satellites/%d", nagiosPath, satelliteID)

	if _, err := os.Stat(satPath); !os.IsNotExist(err) {
		t, err := TarPackEncode(satPath, use64)
		if err != nil {
			log.Println(err)
			return ""
		}
		return t
	}
	log.Println(satPath, "does not exist")
	return ""
}

// ToMap converts a struct to a map using the struct's tags.
//
// ToMap uses tags on struct fields to decide which fields to add to the
// returned map.
func ToMap(in interface{}, tag string) (map[string]interface{}, error) {
	out := make(map[string]interface{})

	v := reflect.ValueOf(in)
	if v.Kind() == reflect.Ptr {
		v = v.Elem()
	}

	// we only accept structs
	if v.Kind() != reflect.Struct {
		return nil, fmt.Errorf("ToMap only accepts structs; got %T", v)
	}

	typ := v.Type()
	for i := 0; i < v.NumField(); i++ {
		// gets us a StructField
		fi := typ.Field(i)
		if tagv := fi.Tag.Get(tag); tagv != "" {
			// set key of map to value in struct field
			out[tagv] = v.Field(i).Interface()
		}
	}
	return out, nil
}
