package sshwrapper

import (
	"bytes"
	"context"
	"fmt"
	"log"
	"os/exec"
	"path/filepath"
	"strings"
	"time"
)

type SSHCommandBackend struct {
	ctx     context.Context
	timeout time.Duration
	conn    *Conn
}

func NewSSHCommandBackend(ctx context.Context, conn *Conn, timeout time.Duration) *SSHCommandBackend {
	return &SSHCommandBackend{
		ctx:     ctx,
		timeout: timeout,
		conn:    conn,
	}
}

func (l *SSHCommandBackend) command(args []string, env map[string]string) []string {
	base := l.conn.commandBase()
	e := make([]string, 0)

	for k, v := range env {
		e = append(e, fmt.Sprintf("%s=%s", k, v))
	}

	cmd := append(
		base,
		e...,
	)
	return append(
		cmd,
		args...,
	)
}

func (l *SSHCommandBackend) RunCommandRaw(host string, directArgs []string, env map[string]string, stdin string) (string, error) {
	ctx, cancel := context.WithTimeout(l.ctx, l.timeout)
	defer cancel()

	args := l.command(directArgs, env)
	c := exec.CommandContext(ctx, args[0], args[1:]...)
	e := make([]string, 0)

	for k, v := range env {
		e = append(e, fmt.Sprintf("%s=%s", k, v))
	}

	c.Env = e

	finished := make(chan error)
	output := &bytes.Buffer{}

	c.Stdin = strings.NewReader(stdin)
	c.Stdout = output
	c.Stderr = output

	cmdArgs := strings.Join(args, " ")

	log.Println(&Error{Name: host, Message: fmt.Sprint("[", cmdArgs, "]")})
	err := c.Start()
	if err != nil {
		return "", &Error{Name: host, Message: fmt.Sprint(cmdArgs, " => could not start process"), Err: err}
	}

	go func() {
		err := c.Wait()
		select {
		case <-ctx.Done():
			return
		default:
			finished <- err
		}
	}()

	select {
	case err := <-finished:
		out := output.String()
		if err != nil {
			return out, &Error{Name: host, Message: fmt.Sprint(cmdArgs, " => \"", out, "\""), Err: err}
		}
		return out, nil
	case <-ctx.Done():
		return "", &Error{Name: host, Message: fmt.Sprint(cmdArgs, " => process timed out")}
	}
}

func (l *SSHCommandBackend) RunCommand(hostuuid string, directArgs []string, env map[string]string) (string, error) {
	return l.RunCommandRaw(fmt.Sprint(l.conn.clientName, ":", hostuuid), directArgs, env, "")
}

func (l *SSHCommandBackend) WriteFile(hostuuid, path string, content []byte) error {
	remoteDir := filepath.Dir(path)
	log.Println("create remote path:", remoteDir)
	if _, err := l.RunCommandRaw(fmt.Sprint(l.conn.clientName, ":", hostuuid), []string{"mkdir", "-p", remoteDir}, map[string]string{}, ""); err != nil {
		return err
	}

	if _, err := l.RunCommandRaw(fmt.Sprint(l.conn.clientName, ":", hostuuid), []string{"bash", "-c", fmt.Sprint("cat > ", path)}, map[string]string{}, string(content)); err != nil {
		return err
	}

	return nil
}
