package sshwrapper

import (
	"bytes"
	"context"
	"fmt"
	"log"
	"net"
	"os"
	"os/exec"
	"strings"
	"sync"
	"sync/atomic"
	"time"
)

// ConnectionTest tries to open a tcp connection
func ConnectionTest(host string, port int) bool {
	timeout := time.Second
	conn, err := net.DialTimeout("tcp", fmt.Sprintf("%s:%d", host, port), timeout)
	if err != nil {
		return false
	}
	if conn != nil {
		defer conn.Close()
		return true
	}
	return false
}

// ConnOpts are used for the ssh connection
type ConnOpts struct {
	Hostname       string
	Port           int
	Login          string
	PrivateKeyPath string
	RemotePort     int
	// Reconnect is the interval time for the connection check, default 5 seconds
	Reconnect time.Duration
}

// Conn manages the ssh connection
type Conn struct {
	ctx        context.Context
	ctxcancel  context.CancelFunc
	opts       ConnOpts
	clientName string

	mtx     sync.Mutex
	wg      sync.WaitGroup
	stopped bool
	psState int64

	cmd    *exec.Cmd
	output *bytes.Buffer

	port chan<- int
}

func dynamicPort() (int, error) {
	addr, err := net.ResolveTCPAddr("tcp", "localhost:0")
	if err != nil {
		return 0, err
	}

	l, err := net.ListenTCP("tcp", addr)
	if err != nil {
		return 0, err
	}
	defer l.Close()
	return l.Addr().(*net.TCPAddr).Port, nil
}

func (c *Conn) commandBase() []string {
	path, err := exec.LookPath("ssh")
	if err != nil {
		log.Panicln(err)
	}
	return []string{
		path,
		"-l",
		c.opts.Login,
		"-o",
		fmt.Sprintf("IdentityFile=%s", c.opts.PrivateKeyPath),
		"-o",
		"StrictHostKeyChecking=accept-new",
		"-o",
		"ControlMaster=no",
		"-o",
		"ControlPath=none",
		"-o",
		"ControlPersist=no",
		"-o",
		"ServerAliveInterval=60",
		"-p",
		fmt.Sprintf("%d", c.opts.Port),
		c.opts.Hostname,
	}

}

// Command returns the full ssh command to run
func (c *Conn) command(localPort int) []string {
	args := c.commandBase()
	return append(args[:len(args)-1], []string{
		"-L",
		fmt.Sprintf("%d:127.0.0.1:%d", localPort, c.opts.RemotePort),
		"-N",
		args[len(args)-1],
	}...)
}

// Close stops the ssh process
func (c *Conn) Close() {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if !c.stopped {
		log.Println(&Error{Name: c.clientName, Message: "stop ssh process"})
		c.stopped = true
		c.ctxcancel()
		if c.cmd != nil {
			if err := c.cmd.Process.Kill(); err != nil {
				log.Println(&Error{Name: c.clientName, Message: "could not stop process", Err: err})
			}
			for atomic.LoadInt64(&c.psState) != 2 {
				time.Sleep(time.Millisecond)
			}
		}
	}
}

func (c *Conn) internalRun() {
	atomic.StoreInt64(&c.psState, 1)
	defer atomic.StoreInt64(&c.psState, 2)

	c.port <- 0

	port, err := dynamicPort()
	if err != nil {
		log.Println(&Error{Name: c.clientName, Message: "could not get a dynamic port for port forwarding", Err: err})
		return
	}
	cmdArgs := c.command(port)
	log.Println(&Error{Name: c.clientName, Message: fmt.Sprint(cmdArgs)})
	c.cmd = exec.CommandContext(c.ctx, cmdArgs[0], cmdArgs[1:]...)
	c.cmd.Stdin = strings.NewReader("")
	c.output = new(bytes.Buffer)
	c.cmd.Stdout = c.output
	c.cmd.Stderr = c.output

	if err := c.cmd.Start(); err != nil {
		log.Println(&Error{Name: c.clientName, Message: "could not start process", Err: err})
		return
	}

	subctx, subcancel := context.WithCancel(c.ctx)

	// We at least try to only send the new port information, when the connection test was successful
	// If we reach the timeout we send it anyways to prevent locks
	go func() {
		tick := time.NewTicker(time.Millisecond * 200)
		defer tick.Stop()
		for {
			select {
			case <-subctx.Done():
				return
			case <-tick.C:
				if ConnectionTest("127.0.0.1", port) {
					select {
					case <-subctx.Done():
						return
					case c.port <- port:
						return
					}
				}
			}
		}
	}()

	c.cmd.Wait()
	subcancel()
}

func (c *Conn) runProcess() {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if !c.stopped {
		state := atomic.LoadInt64(&c.psState)
		switch state {
		case 0:
			// Process was never running
			atomic.StoreInt64(&c.psState, 1) // ensure psState is not 0, before unlock
			go c.internalRun()
		case 1:
			// Process is running
			return
		case 2:
			// Process stopped
			log.Println(&Error{Name: c.clientName, Message: fmt.Sprintf("stopped unexpectedly with output: %s", c.output.String())})
			go c.internalRun()
		}
	}
}

// Run creates the ssh link and should be called by go Run()
// this functions retries the ssh connection after 5 seconds
func (c *Conn) Run(port chan<- int) {
	c.port = port

	timer := time.NewTicker(c.opts.Reconnect)
	defer timer.Stop()
	c.runProcess()

	for {
		select {
		case <-c.ctx.Done():
			if !c.stopped {
				c.Close()
			}
			return
		case <-timer.C:
			c.runProcess()
		}
	}
}

// New creates a new ssh Conn object
func New(ctx context.Context, clientName string, opts ConnOpts) (*Conn, error) {
	if opts.Login == "" || opts.Port == 0 || opts.RemotePort == 0 || opts.PrivateKeyPath == "" ||
		opts.Hostname == "" {
		return nil, &Error{Name: clientName, Message: "missing required ssh connection options"}
	}
	if opts.Reconnect == 0 {
		opts.Reconnect = time.Second * 5
	}

	if _, err := os.Stat(opts.PrivateKeyPath); os.IsNotExist(err) {
		return nil, &Error{Name: clientName, Message: fmt.Sprintf("private key %s does not exist or isn't readable", opts.PrivateKeyPath)}
	}
	ctx, cancel := context.WithCancel(ctx)
	c := &Conn{
		ctx:        ctx,
		ctxcancel:  cancel,
		opts:       opts,
		clientName: clientName,
		output:     nil,
		stopped:    false,
		psState:    0,
	}
	c.wg.Add(1)
	return c, nil
}
