package transfer

import (
	"context"
	"fmt"
	"log"
	"sync"
	"time"

	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpclient"
	"github.com/it-novum/nsta/nstautil"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

// PullSatellite is a satellite pulled via http requests from master
type PullSatellite struct {
	ctx       context.Context
	ctxcancel context.CancelFunc
	mtx       sync.Mutex
	wg        sync.WaitGroup
	stopped   bool

	statusTracker *satstatus.StatusTracker
	db            *oitcdb.Satellite

	httpClient *httpclient.NSTAHTTPClient
	lastExport time.Time
	toMaster   chan<- *gearman.Message
	toSat      gearman.SyncMessageFIFO
	nagiosPath string
}

// SatelliteID returns the db id of the satellite
func (s *PullSatellite) SatelliteID() int {
	return s.db.ID
}

func (s *PullSatellite) SatelliteName() string {
	return s.db.Name
}

// APIKey returns the db signature api key
func (s *PullSatellite) APIKey() string {
	return s.db.APIKey
}

func (s *PullSatellite) LastExport() time.Time {
	return s.lastExport
}

func (s *PullSatellite) SendMessage(msg *gearman.Message) {
	s.toSat.Push(msg)
}

func (s *PullSatellite) Run() {
	timeoutCheck := 300 * time.Second
	timeoutDB := time.Duration(int64(s.db.Timeout)) * time.Second
	if timeoutDB < 240*time.Second {
		timeoutCheck = 60 * time.Second
	} else if timeoutDB < 3600*time.Second {
		timeoutCheck = timeoutDB / 4
	} else {
		timeoutCheck = 3600 * time.Second
	}

	ts := time.NewTicker(time.Second * time.Duration(timeoutCheck))
	ticker := time.NewTicker(time.Second * time.Duration(s.db.Interval))

	nextTs := time.Time{}

	defer func() {
		log.Println(&Error{Name: s.db.Name, Message: "shutting down"})
		ticker.Stop()
		ts.Stop()
		log.Println(&Error{Name: s.db.Name, Message: "close message fifo"})
		s.toSat.Close()
		log.Println(&Error{Name: s.db.Name, Message: "shutdown complete"})
		s.wg.Done()
	}()

	lastSeen := time.Time{}

	for {
		select {
		case <-s.ctx.Done():
			return
		case <-ts.C:
			absTimeout := lastSeen.Add(timeoutDB)
			if !absTimeout.After(time.Now()) {
				log.Println(&Error{Name: s.db.Name, Message: fmt.Sprintf("Last Seen: %s, Timeout: %s", lastSeen.Format(time.RFC3339), absTimeout.Format(time.RFC3339))})
				s.statusTracker.Update <- &satstatus.Status{
					SatelliteID: s.db.ID,
					Status:      oitcdb.SatelliteStatusFailed,
					Err:         &Error{Name: s.db.Name, Message: "timeout reached"},
				}
			}
		case <-ticker.C:
			status := &satstatus.Status{
				SatelliteID: s.db.ID,
				Status:      oitcdb.SatelliteStatusOK,
			}
			lastExport := s.LastExport()
			msgs := s.toSat.GetAll(8192)
			origMsgs := msgs
			// we do not queue the timesync message, because it should be sent to the client immediatly
			if s.db.UseTimesync {
				now := time.Now()
				if now.After(nextTs) {
					nextTs = now.Add(time.Minute)
					ts := gearman.TimesyncMessage{
						SatelliteID: s.db.ID,
						SetTime:     now,
					}
					if data, err := json.Marshal(ts); err != nil {
						log.Println(&Error{Name: s.db.Name, Message: "could not create timesync message", Err: err})
					} else {
						msgs = append(msgs, &gearman.Message{
							Type:   gearman.MessageTimesync,
							String: string(data),
						})
					}
				}
			}

			response, err := s.httpClient.SendMessages(msgs, lastExport)
			if err != nil {
				status.Err = &Error{Name: s.db.Name, Message: "could not send/receive messages", Err: err}
				log.Println(status.Err)
				if len(origMsgs) > 0 {
					s.toSat.PushAll(origMsgs, true)
				}
			}
			if response != nil {
				lastSeen = time.Now()
				status.LastSeen = lastSeen
				if response.SatelliteID != s.db.ID || response.APIKey != s.db.APIKey {
					log.Println(&Error{Name: s.db.Name, Message: fmt.Sprint("sent ", len(msgs), " messages")})
					status.Err = &Error{Name: s.db.Name, Message: "authentication of response failed"}
					log.Println(status.Err)
				} else {
					log.Println(&Error{Name: s.db.Name, Message: fmt.Sprint("sent ", len(msgs), " and received ", len(response.Messages), " messages")})
					for _, msg := range response.Messages {
						s.toMaster <- msg
					}
					if s.LastExport().After(response.LastExport) {
						tData := nstautil.CreateExportTar(s.nagiosPath, s.db.ID, true)
						if tData != "" {
							if exportResponse, err := s.httpClient.ExportExchange(lastExport, tData); err != nil {
								log.Println(err)
								status.Err = err
							} else {
								status.LastExport = exportResponse.LastExport
							}
						}
					}
				}
			} else if err != nil {
				status.Status = oitcdb.SatelliteStatusFailed
				log.Println(&Error{Name: s.db.Name, Message: fmt.Sprint("sent ", len(msgs), " messages")})
			}
			s.statusTracker.Update <- status
		}
	}
}

func (s *PullSatellite) Close() {
	s.mtx.Lock()
	defer s.mtx.Unlock()

	if !s.stopped {
		s.stopped = true
		s.ctxcancel()
	}
	s.wg.Wait()
}

func (s *PullSatellite) Compare(dbSat *oitcdb.Satellite) bool {
	return s.db.Compare(dbSat)
}

func (s *PullSatellite) Reload() {
	if !s.db.NSTASyncInstance {
		log.Println(&Error{Name: s.db.Name, Message: "skip export"})
		return
	}
	s.lastExport = time.Now()
}

func newPullSatellite(ctx context.Context, satOpts satelliteOptions) (*PullSatellite, error) {
	ctx, cancel := context.WithCancel(ctx)
	proxy := ""
	if satOpts.db.UseProxy > 0 {
		proxy = satOpts.db.ProxyURL
	}
	httpClient, err := httpclient.New(ctx, satOpts.db.URL, satOpts.db.ID, satOpts.db.APIKey, proxy, satOpts.db.VerifyCertificate)
	if err != nil {
		return nil, err
	}
	s := &PullSatellite{
		ctx:           ctx,
		ctxcancel:     cancel,
		statusTracker: satOpts.statusTracker,
		db:            satOpts.db,
		httpClient:    httpClient,
		lastExport:    time.Now(),
		toMaster:      satOpts.toMaster,
		nagiosPath:    satOpts.nagiosPath,
	}
	s.wg.Add(1)
	s.toSat.Start()

	return s, nil
}
