package transfer

import (
	"context"
	"fmt"
	"log"
	"sync"
	"time"

	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpproto"
	"github.com/it-novum/nsta/httpserver"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
)

type PushSatellite struct {
	ctx       context.Context
	ctxcancel context.CancelFunc
	mtx       sync.Mutex
	wg        sync.WaitGroup
	stopped   bool

	statusTracker *satstatus.StatusTracker
	db            *oitcdb.Satellite

	lastExport time.Time
	nagiosPath string
	toMaster   chan<- *gearman.Message
	toSat      gearman.SyncMessageFIFO

	io httpserver.RegisterSatelliteIO

	nextTs   time.Time
	lastSeen time.Time
}

// SatelliteID returns the db id of the satellite
func (s *PushSatellite) SatelliteID() int {
	return s.db.ID
}

// SatelliteName returns the db name of the satellite
func (s *PushSatellite) SatelliteName() string {
	return s.db.Name
}

// APIKey returns the db signature api key
func (s *PushSatellite) APIKey() string {
	return s.db.APIKey
}

func (s *PushSatellite) Incoming(msg *gearman.Message) {
	s.toMaster <- msg
}

func (s *PushSatellite) Outgoing() []*gearman.Message {
	msgs := s.toSat.GetAll(8192)
	now := time.Now()
	// We insert the timesync message on demand, because we don't want to sync a totally outdated timestamp
	if s.db.UseTimesync && now.After(s.nextTs) {
		s.nextTs = now.Add(time.Minute)
		ts := gearman.TimesyncMessage{
			SatelliteID: s.db.ID,
			SetTime:     now,
		}
		if data, err := json.Marshal(ts); err != nil {
			log.Println(&Error{Name: s.db.Name, Message: "could not create timesync message", Err: err})
		} else {
			msgs = append(msgs, &gearman.Message{
				Type:   gearman.MessageTimesync,
				String: string(data),
			})
		}
	}
	return msgs
}

func (s *PushSatellite) LastExport() time.Time {
	return s.lastExport
}

func (s *PushSatellite) SendMessage(msg *gearman.Message) {
	s.toSat.Push(msg)
}

func (s *PushSatellite) Export(msg *httpproto.ExportData) error {
	return nil
}

func (s *PushSatellite) WebStatus(status *satstatus.Status) {
	if !status.LastSeen.IsZero() {
		s.lastSeen = status.LastSeen
	}
	s.statusTracker.Update <- status
}

// Run is the main loop of this satellite
func (s *PushSatellite) Run() {
	timeoutCheck := 300 * time.Second
	timeoutDB := time.Duration(int64(s.db.Timeout)) * time.Second
	if timeoutDB < 240*time.Second {
		timeoutCheck = 60 * time.Second
	} else if timeoutDB < 3600*time.Second {
		timeoutCheck = timeoutDB / 4
	} else {
		timeoutCheck = 3600 * time.Second
	}

	ts := time.NewTicker(timeoutCheck)

	defer func() {
		log.Println(&Error{Name: s.db.Name, Message: "shutting down"})
		s.io.UnregisterSatelliteIO(s.db.ID)
		ts.Stop()
		log.Println(&Error{Name: s.db.Name, Message: "close message fifo"})
		s.toSat.Close()
		log.Println(&Error{Name: s.db.Name, Message: "shutdown complete"})
		s.wg.Done()
	}()
	for {
		select {
		case <-s.ctx.Done():
			return
		case <-ts.C:
			absTimeout := s.lastSeen.Add(timeoutDB)
			if !absTimeout.After(time.Now()) {
				log.Println(&Error{Name: s.db.Name, Message: fmt.Sprintf("Last Seen: %s, Timeout: %s", s.lastSeen.Format(time.RFC3339), absTimeout.Format(time.RFC3339))})
				s.statusTracker.Update <- &satstatus.Status{
					SatelliteID: s.db.ID,
					Status:      oitcdb.SatelliteStatusFailed,
					Err:         &Error{Name: s.db.Name, Message: "timeout reached"},
				}
			}
		}
	}
}

// Close stops the satellite
func (s *PushSatellite) Close() {
	s.mtx.Lock()
	defer s.mtx.Unlock()

	if !s.stopped {
		s.stopped = true
		s.ctxcancel()
	}
	s.wg.Wait()
}

// Compare checks if the satellite matches the given db satellite data
func (s *PushSatellite) Compare(dbSat *oitcdb.Satellite) bool {
	return s.db.Compare(dbSat)
}

func (s *PushSatellite) Reload() {
	if !s.db.NSTASyncInstance {
		log.Println(&Error{Name: s.db.Name, Message: "skip export"})
		return
	}
	s.lastExport = time.Now()
}

func newPushSatellite(ctx context.Context, satOpts satelliteOptions) (*PushSatellite, error) {
	ctx, cancel := context.WithCancel(ctx)
	s := &PushSatellite{
		ctx:           ctx,
		ctxcancel:     cancel,
		statusTracker: satOpts.statusTracker,
		db:            satOpts.db,
		nagiosPath:    satOpts.nagiosPath,
		io:            satOpts.io,
		lastExport:    time.Now(),
		toMaster:      satOpts.toMaster,
	}
	satOpts.io.RegisterSatelliteIO(s)
	s.toSat.Start()
	s.wg.Add(1)
	return s, nil
}
