package transfer

import (
	"context"
	"fmt"
	"io"
	"log"
	"os"
	"os/signal"
	"sync"
	"syscall"
	"time"

	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpserver"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
	"github.com/it-novum/nsta/sshwrapper"
)

type satelliteOptions struct {
	//all
	statusTracker *satstatus.StatusTracker
	db            *oitcdb.Satellite

	nagiosPath string

	toMaster chan<- *gearman.Message

	// ssh
	sshOpts sshwrapper.ConnOpts

	// http_push
	io httpserver.RegisterSatelliteIO
}

// Satellite is the common interface of all different satellite types
type Satellite interface {
	SatelliteID() int
	SatelliteName() string
	Run()
	Close()
	Compare(*oitcdb.Satellite) bool
	SendMessage(msg *gearman.Message)
	Reload()
}

// closeSatellites is a helper that closes all satellites
func closeSatellites(satellites []Satellite) {
	if len(satellites) > 0 {
		var wg sync.WaitGroup
		wg.Add(len(satellites))
		for _, sat := range satellites {
			go func(sat Satellite) {
				sat.Close()
				wg.Done()
			}(sat)
		}
		wg.Wait()
	}
}

// MasterSystem handles the master to satellite connections and incoming and outgoing messages
type MasterSystem struct {
	ctx  context.Context
	oitc *oitcdb.OITC

	webserver     *httpserver.NSTAHTTPServer
	localGearman  *gearman.Client
	address       string
	satellites    map[int]Satellite
	toLocal       gearman.SyncMessageFIFO
	toSatFan      gearman.SyncMessageFIFO
	nagiosPath    string
	statusTracker *satstatus.StatusTracker
}

// Run starts the actual work here
func (m *MasterSystem) Run() {
	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM, syscall.SIGHUP)

	go m.localGearman.Run()
	m.reload()
	m.webserver.Run()

	defer func() {
		m.webserver.Stop()
		sats := make([]Satellite, 0)
		for _, sat := range m.satellites {
			sats = append(sats, sat)
		}
		closeSatellites(sats)
		m.toLocal.Close()
		m.toSatFan.Close()
		m.localGearman.Close()
		m.statusTracker.Close()
		m.oitc.Close()
	}()

	for {
		select {
		case <-m.ctx.Done():
			return
		case sig := <-sigs:
			log.Println(&Error{Name: "Master", Message: fmt.Sprint("received signal:", sig.String())})
			if sig.String() == "hangup" {
				// reload
				m.reload()
			} else {
				return
			}
		case <-m.toSatFan.Notify():
			msg := m.toSatFan.Get()
			if msg != nil {
				satID := msg.Content().SatelliteID
				sat, ok := m.satellites[satID]
				if ok {
					sat.SendMessage(msg)
				} else {
					log.Println(&Error{Name: "Master", Message: fmt.Sprint("unknown satellite id: ", satID)})
				}
			}
		case <-m.toLocal.Notify():
			msgs := m.toLocal.GetAll(8192)
			for _, msg := range msgs {
				m.localGearman.SendMessage(msg)
			}
		}
	}
}

func (m *MasterSystem) reload() {
	log.Println(&Error{Name: "Local", Message: "reload of satellite configuration started"})
	dbSats := m.oitc.FetchSatellites()

	newSatMap := make(map[int]Satellite)
	keepSats := []Satellite{}
	dropSats := []Satellite{}
	newDbSats := []*oitcdb.Satellite{}

	for satID, sat := range m.satellites {
		keep := false
	dbloop:
		for _, dbSat := range dbSats {
			if sat.Compare(dbSat) {
				keep = true
				break dbloop
			}
		}
		if keep {
			keepSats = append(keepSats, sat)
			newSatMap[satID] = sat
			log.Println(&Error{Name: sat.SatelliteName(), Message: "satellite configuration not changed -> keep running"})
		} else {
			log.Println(&Error{Name: sat.SatelliteName(), Message: "satellite configuration changed or deleted -> close"})
			dropSats = append(dropSats, sat)
		}
	}

	for _, dbSat := range dbSats {
		exists := false
	keeploop:
		for _, sat := range keepSats {
			if sat.Compare(dbSat) {
				exists = true
				break keeploop
			}
		}
		if !exists {
			log.Println(&Error{Name: dbSat.Name, Message: "satellite configuration new -> start"})
			newDbSats = append(newDbSats, dbSat)
		}
	}

	beginReload := time.Now()
	closeSatellites(dropSats)

	newSats := []Satellite{}
	for _, dbSat := range newDbSats {
		opts := satelliteOptions{
			statusTracker: m.statusTracker,
			db:            dbSat,
			nagiosPath:    m.nagiosPath,
			toMaster:      m.toLocal.Direct(),
			sshOpts: sshwrapper.ConnOpts{
				Hostname:       dbSat.Address,
				Login:          dbSat.Login,
				Port:           dbSat.Port,
				PrivateKeyPath: dbSat.PrivateKeyPath,
				RemotePort:     dbSat.RemotePort,
			},
			io: m.webserver,
		}
		if dbSat.SyncMethod == "ssh" {
			sat, err := newSSHSatellite(m.ctx, opts)
			if err != nil {
				log.Println(&Error{Name: dbSat.Name, Message: "could not create ssh client", Err: err})
			} else {
				newSats = append(newSats, sat)
				newSatMap[dbSat.ID] = sat
			}
		} else if dbSat.SyncMethod == "https_pull" {
			sat, err := newPullSatellite(m.ctx, opts)
			if err != nil {
				log.Println(&Error{Name: dbSat.Name, Message: "could not create https pull client", Err: err})
			} else {
				newSats = append(newSats, sat)
				newSatMap[dbSat.ID] = sat
			}
		} else if dbSat.SyncMethod == "https_push" {
			sat, err := newPushSatellite(m.ctx, opts)
			if err != nil {
				log.Println(&Error{Name: dbSat.Name, Message: "could not create https push server", Err: err})
			} else {
				newSats = append(newSats, sat)
				newSatMap[dbSat.ID] = sat
			}
		}
	}

	for _, sat := range newSats {
		go sat.Run()
	}

	satList := make([]satstatus.Satellite, 0)
	for _, sat := range newSatMap {
		satList = append(satList, sat)
	}
	m.statusTracker.Reload(satList, beginReload)

	wg := sync.WaitGroup{}
	wg.Add(len(newSatMap))
	for _, sat := range newSatMap {
		go func(ls Satellite) {
			ls.Reload()
			wg.Done()
		}(sat)
	}
	wg.Wait()

	m.satellites = newSatMap
	log.Println(&Error{Name: "Local", Message: "reload of satellite configuration finished"})
}

func New(ctx context.Context, accessLogWriter io.Writer, mycnf, gearmanAddress, webserverAddressHTTP, webserverAddressHTTPS, certFile, keyFile, nagiosPath string) (*MasterSystem, error) {
	var err error
	oitc, err := oitcdb.New(mycnf)
	if err != nil {
		log.Fatalln(err)
	}

	m := &MasterSystem{
		ctx:           ctx,
		webserver:     httpserver.New(accessLogWriter, nagiosPath, webserverAddressHTTP, webserverAddressHTTPS, certFile, keyFile, httpserver.WebserverTypeMaster),
		nagiosPath:    nagiosPath,
		oitc:          oitc,
		address:       gearmanAddress,
		satellites:    make(map[int]Satellite, 0),
		statusTracker: satstatus.New(oitc),
	}
	m.toLocal.Start()
	m.toSatFan.Start()

	m.localGearman = gearman.NewMaster(ctx, "Local", gearmanAddress, m.toSatFan.Direct())

	return m, nil
}
