package transfer

import (
	"context"
	"fmt"
	"log"
	"path/filepath"
	"sync"
	"time"

	"github.com/it-novum/nsta/checkmk"
	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/nstautil"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
	"github.com/it-novum/nsta/sshwrapper"
)

// SSHSatellite manages the connection and message exchange to a satellite via ssh
type SSHSatellite struct {
	ctx       context.Context
	ctxcancel context.CancelFunc
	wg        sync.WaitGroup
	reloadC   chan time.Time
	portChan  chan int
	mtx       sync.Mutex
	stopped   bool

	statusTracker *satstatus.StatusTracker
	db            *oitcdb.Satellite

	lastExport time.Time

	satGearman       *gearman.Client
	ssh              *sshwrapper.Conn
	sshOpts          sshwrapper.ConnOpts
	nagiosPath       string
	nagiosConfigPath string

	toMaster chan<- *gearman.Message
	toSat    gearman.SyncMessageFIFO

	cmk        checkmk.Checkmk
	cmdBackend *sshwrapper.SSHCommandBackend
}

func (s *SSHSatellite) SatelliteID() int {
	return s.db.ID
}

func (s *SSHSatellite) SatelliteName() string {
	return s.db.Name
}

// Compare checks if the satellite matches the database config
func (s *SSHSatellite) Compare(dbSat *oitcdb.Satellite) bool {
	return s.db.Compare(dbSat)
}

// SendMessage to the satellite
func (s *SSHSatellite) SendMessage(msg *gearman.Message) {
	s.toSat.Push(msg)
}

func (s *SSHSatellite) resetPort(port int) {
	status := &satstatus.Status{
		SatelliteID: s.db.ID,
		Status:      oitcdb.SatelliteStatusFailed,
	}
	s.statusTracker.Update <- status
	if s.satGearman != nil {
		s.satGearman.Close()
		s.satGearman = nil
	}
	if port != 0 {
		s.satGearman = gearman.NewSatellite(s.ctx, s.db.Name, fmt.Sprintf("127.0.0.1:%d", port), s.toMaster)
		go s.satGearman.Run()
		status.Status = oitcdb.SatelliteStatusOK
		status.LastSeen = time.Now()
	}
}

func (s *SSHSatellite) recheckStatus() {
	status := &satstatus.Status{
		SatelliteID: s.db.ID,
		Status:      oitcdb.SatelliteStatusFailed,
	}
	if s.satGearman != nil && s.satGearman.IsRunning() {
		status.Status = oitcdb.SatelliteStatusOK
	}
	s.statusTracker.Update <- status
}

// Run is the background go worker
func (s *SSHSatellite) Run() {
	ts := time.NewTicker(time.Minute)
	statusT := time.NewTicker(time.Second * 10)

	if !s.db.UseTimesync {
		ts.Stop()
	}

	defer func() {
		log.Println(&Error{Name: s.db.Name, Message: "shutting down"})
		ts.Stop()
		statusT.Stop()
		if s.satGearman != nil {
			log.Println(&Error{Name: s.db.Name, Message: "close gearman connection"})
			s.satGearman.Close()
			s.satGearman = nil
		}
		log.Println(&Error{Name: s.db.Name, Message: "stop ssh connection"})
		s.ssh.Close()
		log.Println(&Error{Name: s.db.Name, Message: "close message fifo"})
		s.toSat.Close()
		log.Println(&Error{Name: s.db.Name, Message: "shutdown complete"})
		s.wg.Done()
	}()

	for {
		if s.satGearman == nil {
			select {
			case <-s.ctx.Done():
				return
			case <-statusT.C:
				s.recheckStatus()
			case port := <-s.portChan:
				s.resetPort(port)
			}
		} else {
			select {
			case <-s.ctx.Done():
				return
			case <-statusT.C:
				s.recheckStatus()
			case port := <-s.portChan:
				s.resetPort(port)
			case <-s.toSat.Notify():
				msg := s.toSat.Get()
				if msg != nil {
					if msg.Type == gearman.MessageCheckmk {
						go func() {
							response, err := s.cmk.ProcessMessage(s.cmdBackend, msg).Serialize()
							if err != nil {
								log.Println(&Error{Name: s.db.Name, Message: "could no serialize checkmk answer:", Err: err})
							}
							s.toMaster <- &gearman.Message{
								Type:   gearman.MessageCheckmk,
								String: response,
							}
						}()
					} else {
						s.satGearman.SendMessage(msg)
					}
				}
			case <-ts.C:
				if s.db.UseTimesync {
					go func() {
						if _, err := s.cmdBackend.RunCommandRaw(s.db.Name, []string{"sudo", "date", "-s", fmt.Sprint("@", time.Now().Unix())}, map[string]string{}, ""); err != nil {
							log.Println(err)
						} else {
							log.Println(&Error{Name: s.db.Name, Message: "time update was successful"})
						}
					}()
				}
			}
		}
	}
}

// Close stops the satellite connection
func (s *SSHSatellite) Close() {
	s.mtx.Lock()
	defer s.mtx.Unlock()

	if !s.stopped {
		s.stopped = true
		s.ctxcancel()
	}
	s.wg.Wait()
}

// Reload implements a process reload signal handler
func (s *SSHSatellite) Reload() {
	if !s.db.NSTASyncInstance {
		log.Println(&Error{Name: s.db.Name, Message: "skip export"})
		return
	}
	s.lastExport = time.Now()
	status := &satstatus.Status{
		SatelliteID: s.db.ID,
	}
	defer func() {
		s.statusTracker.Update <- status
	}()
	log.Println(&Error{Name: s.db.Name, Message: "copy configuration"})
	tarData := nstautil.CreateExportTar(s.nagiosPath, s.db.ID, false)
	if tarData != "" {
		if _, err := s.cmdBackend.RunCommandRaw(s.db.Name, []string{"tar", "xz", "-C", s.nagiosConfigPath}, map[string]string{}, tarData); err != nil {
			status.Err = &Error{Name: s.db.Name, Message: "export failed", Err: err}
			fmt.Println(status.Err)
			return
		}
		log.Println(&Error{Name: s.db.Name, Message: "restart monitoring"})
		if _, err := s.cmdBackend.RunCommandRaw(s.db.Name, []string{"sudo", "/opt/openitc/nagios/bin/restart-monitoring.sh"}, map[string]string{}, ""); err != nil {
			status.Err = &Error{Name: s.db.Name, Message: "export failed", Err: err}
			fmt.Println(status.Err)
			return
		}
		status.LastExport = time.Now()
		status.LastSeen = time.Now()
		log.Println(&Error{Name: s.db.Name, Message: "export successful"})
	} else {
		status.Err = &Error{Name: s.db.Name, Message: "could not create export archive"}
		log.Println(status.Err)
	}
}

// NewSatellite initializes an SSHSatallite
func newSSHSatellite(ctx context.Context, satOpts satelliteOptions) (*SSHSatellite, error) {
	port := make(chan int)

	ctx, cancel := context.WithCancel(ctx)
	_ = cancel // Fix for linter
	sshc, err := sshwrapper.New(ctx, satOpts.db.Name, satOpts.sshOpts)
	if err != nil {
		return nil, err
	}
	go sshc.Run(port)

	s := &SSHSatellite{
		ctx:              ctx,
		ctxcancel:        cancel,
		statusTracker:    satOpts.statusTracker,
		db:               satOpts.db,
		ssh:              sshc,
		sshOpts:          satOpts.sshOpts,
		toMaster:         satOpts.toMaster,
		nagiosPath:       satOpts.nagiosPath,
		nagiosConfigPath: filepath.Join(satOpts.nagiosPath, "etc"),
		portChan:         port,
		stopped:          false,
		lastExport:       time.Now(),
		cmdBackend:       sshwrapper.NewSSHCommandBackend(ctx, sshc, time.Second*120),
	}
	s.cmk.SatelliteName = satOpts.db.Name
	s.toSat.Start()
	s.wg.Add(1)
	return s, nil
}
